package yukihane.inqubus.filewatch;

import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardWatchEventKinds;
import java.nio.file.WatchEvent;
import java.nio.file.WatchEvent.Kind;
import java.nio.file.WatchKey;
import java.nio.file.WatchService;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author yuki
 */
public final class FileWatch implements Runnable {

    private static final Logger logger = Logger.getLogger(FileWatch.class.getName());
    private final Set<Path> files;
    private final Set<Path> directories;

    public FileWatch(Collection<Path> dirs) {
        FileDir res = new FileDir();
        getFileDir(dirs, res);

        this.files = new HashSet<>(res.files);
        this.directories = new HashSet<>(res.dirs);
        if(logger.isLoggable(Level.FINER)) {
            logger.log(Level.FINER, "search dirs: {0}", directories);
            logger.log(Level.FINER, "exist files: {0}", files);
        }
    }

    public Set<Path> getFiles() {
        synchronized (files) {
            return new HashSet<>(files);
        }
    }

    private void getFileDir(Iterable<Path> paths, FileDir result) {
        for (Path p : paths) {
            getFileDir(p, result);
        }
    }

    private void getFileDir(Path path, FileDir result) {
        try {
            if (Files.isDirectory(path)) {
                result.dirs.add(path);
                DirectoryStream<Path> dir = Files.newDirectoryStream(path);
                getFileDir(dir, result);
            } else if (Files.isRegularFile(path)) {
                result.files.add(path);
            }
        } catch (IOException ex) {
            logger.log(Level.WARNING, "パスの処理中にエラー: {0}", new Object[]{path});
        }
    }

    private static class FileDir {

        private final Set<Path> files = new HashSet<>();
        private final Set<Path> dirs = new HashSet<>();
    }

    @Override
    public void run() {
        try {
            final WatchService ws = FileSystems.getDefault().newWatchService();
            for (Path p : directories) {
                p.register(ws, StandardWatchEventKinds.ENTRY_CREATE, StandardWatchEventKinds.ENTRY_DELETE);
            }

            while (true) {
                final WatchKey wk = ws.take();
                for (final WatchEvent<?> event : wk.pollEvents()) {
                    final Kind<?> kind = event.kind();
                    if (kind.equals(StandardWatchEventKinds.ENTRY_CREATE)) {
                        final Path p = (Path) event.context();
                        if (Files.isRegularFile(p)) {
                            logger.log(Level.FINEST, "ファイル追加: {0}", p);
                            synchronized (files) {
                                files.add(p);
                            }
                        }
                    } else if (kind.equals(StandardWatchEventKinds.ENTRY_DELETE)) {
                        final Path p = (Path) event.context();
                        if (Files.isRegularFile(p)) {
                            logger.log(Level.FINEST, "ファイル削除: {0}", p);
                            synchronized (files) {
                                files.remove(p);
                            }
                        }
                    }
                }

                if (!wk.reset()) {
                    System.out.println("No longer valid");
                    wk.cancel();
                    ws.close();
                    break;
                }

            }
        } catch (InterruptedException ex) {
            ex.printStackTrace();
        } catch (IOException ex) {
            ex.printStackTrace();
        }
    }
}
